#!/usr/local/bin/perl
    # William F. Dudley Jr.
    # Feb 1, 2018
    # reads troff file (assuming mm macros) and generates html file
    # handles: .AL, .BL, .ce, .DS/.DE, .ft, .P, .PH, .PF, .sp, .TS/.TE
    # (numbered lists, bulleted lists, centered text, displays, paragraphs,
    # page header, page footer (as subtitle), blank lines, tables)
    # handles font changes .ft and any of R,B,I,BI,H,HB,HI,CW
    # ignores troff comments \", \#, and .ig
    # does not handle lots of stuff, including:
    # \fI, \fR

sub gen_head();
sub set_font($$);
sub usage();

    $main::trfile;
    %main::debug = ();
    for(my $i = 0 ; $i <= $#ARGV ; $i++) {
	if(-s $ARGV[$i]) {
	    $main::trfile = $ARGV[$i];
	}
	elsif($ARGV[$i] =~ /^-d([fht])/) {
	    $main::debug{$1}++;
	}
	else {
	    usage();
	}
    }
    unless(-s $main::trfile) {
	usage();
    }
    open(IN, "<$main::trfile");
    open(OUT, ">$main::trfile.html");
    my $inpara  = 0;
    my $intable = 0;
    my $inlist  = 0;
    my $list_type = "";
    my $t_needs_atr = 0;
    my $t_needs_fmt = 0;
    my %t_fmt = ();
    my $tfmtc = 0;
    my $tfmti = 0;
    my $inrow = 0;
    my $incol = 0;
    my $indisp = 0;
    my $lastfont = "";
    my $currfont = "R";
    my $realfont = "Rbi";
    my $outside_ds_font = "";
    my $incenter = 0;
    my $inignore = "";
    $main::title = "";
    $main::subtitle = "";
    my $head = 0;
    while(<IN>) {
	chomp;
	s/
//;
	s/\\\[u00B2\]/<sup>2<\/sup>/ig;
	s/\\\&//g;	# troff 0 width space
	s/\\["#].*//g;	# troff comment
	s/\\fB/<B>/g;	# bold
	s/\\f[PR]/<\/B>/g;	# not bold
	if(/^\..*/ || $intable) {
	    if(/^\.PH/) {
		$main::debug{'h'} && printf("found $_\n");
		if($_ =~ /"'(.*)'"/) {
		    $main::debug{'h'} && printf(".PH %s\n", $1);
		}
		if(/"'(.*')"/) {
		    $main::debug{'h'} && printf("match .PH %s\n", $1);
		    @fields = split(/'/, $1);
		    $main::debug{'h'} && printf("sizeof fields is %d\n", $#fields + 1);
		    $main::debug{'h'} && printf(".PH %s\n", join(",", @fields));
		    $main::title = join(" \&nbsp; ", @fields);
		    $main::debug{'h'} && printf("title is \"%s\"\n", $main::title);
		}
	    }
	    elsif(/^\.PF/) {
		$main::debug{'h'} && printf("found $_\n");
		if($_ =~ /"'(.*)'"/) {
		    $main::debug{'h'} && printf(".PF %s\n", $1);
		}
		if(/"'(.*')"/) {
		    $main::debug{'h'} && printf("match .PF %s\n", $1);
		    @fields = split(/'/, $1);
		    $main::debug{'h'} && printf("sizeof fields is %d\n", $#fields + 1);
		    $main::debug{'h'} && printf(".PF %s\n", join(",", @fields));
		    $main::subtitle = join(" \&nbsp; ", @fields);
		    $main::debug{'h'} && printf("title is \"%s\"\n", $main::subtitle);
		}
	    }
	    elsif(/^\.ft/) {
		if($_ =~ /^\.ft ([BIRH]+)/) {
		    $lastfont = $currfont;
		    $currfont = $1;
		}
		elsif($_ =~ /^\.ft (CW)/) {
		    $lastfont = $currfont;
		    $currfont = $1;
		}
		elsif($_ =~ /^\.ft (.+)/) {
		    printf("unrecognized font change to %s\n", $1)
		}
		else {
		    my $lf = $currfont;
		    $main::debug{'f'} && printf("font reset to %s\n", $lastfont);
		    $currfont = $lastfont;
		    $lastfont = $lf;
		}
		set_font($currfont, $lastfont);
	    }
	    elsif(/^\.ig/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		$inignore = "..";
		if($_ =~ /^\.ig (.+)/) {
		    $inignore = "." . $1;
		}
	    }
	    elsif($inignore && ($_ eq $inignore)) {
		$inignore = "";
	    }
	    elsif(/^\.ce/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		$incenter = -1;
		if($_ =~ /^\.ce (\d+)/) {
		    $incenter = $1;
		}
		print OUT "<CENTER>\n";
	    }
	    elsif(/^\.sp/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		my $num = 1;
		if($_ =~ /^\.sp (\d+)/) {
		    $num = $1;
		}
		if($_ =~ /^\.sp -.+/) {
		    $num = 0;
		}
		for(my $i = 0 ; $i < $num ; $i++) {
		    print OUT "</BR>\n";
		}
	    }
	    elsif(/^\.P/i) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		print OUT "<P>\n";
		$inpara = 1;
	    }
	    elsif(/^\.DS/) {
		gen_head();
		$outside_ds_font = $currfont;
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		$indisp = "L";
		if($_ =~ /^\.DS ([LCR])/) {
		    $indisp = $1;
		}
		if($indisp eq "C") {
		    print OUT "<CENTER>\n";
		}
	    }
	    elsif(/^\.DE/) {
		if($indisp eq "C") {
		    print OUT "</CENTER>\n";
		}
		if($currfont ne $outside_ds_font) {
		    my $lf = $currfont;
		    $currfont = $outside_ds_font;
		    $lastfont = $lf;
		    set_font($currfont, $lastfont);
		    $main::debug{'f'} && printf(".DE, font reset to %s, lastfont = %s\n", $currfont, $lastfont);
		}
		print OUT "<br/>\n";
		$indisp = 0;
	    }
	    elsif(/^\.BL/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		print OUT "<UL>\n";
		$inlist++;
		$list_type = "UL";
	    }
	    elsif(/^\.AL/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		my $ol_type = "";
		if($_ =~ /^\.AL ([1AaIi])/) {
		    $ol_type = sprintf("type=\"%s\"", $1);
		}
		gen_head();
		printf OUT ("<OL %s>\n", $ol_type);
		$inlist++;
		$list_type = "OL";
	    }
	    elsif(/^\.LE/) {
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		if($inlist) {
		    $inlist--;
		    printf OUT ("</%s>\n", $list_type);
		}
	    }
	    elsif(/^\.LI/) {
		print OUT "<LI>\n";
	    }
	    elsif(/^\.TS/) {
		gen_head();
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		$intable++;
		$t_needs_atr = 1;
	    }
	    elsif($intable && $t_needs_atr && /;$/i) {
		my $center = "";
		my $border = "";
		if(/center/) { $center = "align=\"center\""; }
		if(/allbox/) { $border = "border=\"1\""; }
		if($inpara) {
		    $inpara = 0;
		    print OUT "</P>\n";
		}
		$main::debug{'t'} && print "NEW TABLE ============================ \n";
		printf OUT ("<TABLE %s %s>\n", $center, $border);
		$t_needs_atr = 0;
		$t_needs_fmt = 1;
		%t_fmt = ();
		$tfmtc = 0;
		$inrow = 0;
	    }
	    elsif($intable && $t_needs_fmt && /[lrcns]/i) {
		$t_fmt{$tfmtc++} = $_;
		if(/\.$/) {
		    $t_needs_fmt = 0;
		    $tfmti = 0;
		}
	    }
	    elsif($intable && /^\.TE/) {
		if($inrow) {
		    print OUT "</TR>\n";
		}
		if($intable) {
		    $intable--;
		    print OUT "</TABLE><br/>\n";
		}
	    }
	    elsif($intable) {
		print OUT "    <TR>\n";
		@fields = split(/\t/);
		my $fmt = $t_fmt{$tfmti};
		@fmt = split(/ +/, $fmt);
		if(($tfmti + 1) < $tfmtc) {
		    $tfmti++;
		}
		my $align = "";
		my $span = "";
		my $d = 0;
		$main::debug{'t'} && printf("format for row is %s\n", join(" ", @fmt));
		for(my $i = 0 ; $i <= $#fmt ; $i++, $d++) {
		    my $td = "TD";
		    if($fmt[$i] =~ /B/i) {
			$td = "TH";
		    }
		    if($fmt[$i] =~ /l/) {
			$align = "align=\"left\"";
		    }
		    elsif($fmt[$i] =~ /r/) {
			$align = "align=\"right\"";
		    }
		    elsif($fmt[$i] =~ /n/) {
			$align = "align=\"right\"";
		    }
		    elsif($fmt[$i] =~ /c/) {
			$align = "align=\"center\"";
		    }
		    elsif($fmt[$i] =~ /s/) {
			next;
		    }
		    my $spancols = 0;
		    for($m = $i + 1; $m <= $#fmt ; $m++) {
			if($fmt[$m] =~ /s/i) {
			    $spancols++;
			}
			else { last; }
		    }
		    if($spancols) {
			$span = sprintf("colspan=\"%d\"", $spancols + 1);
			$i+= $spancols;
		    }
		    if($incol) {
			print OUT "        </TD>\n";
		    }
		    $main::debug{'t'} && printf("        <%s %s %s>\n", $td, $align, $span);
		    printf OUT ("        <%s %s %s>\n", $td, $align, $span);
		    printf OUT ("            %s\n", $fields[$d]);
		    $incol = 1;
		}
		if($incol) {
		    print OUT "        </${td}>\n";
		}
		print OUT "    </TR>\n";
		$incol = 0;
	    }
	    @fields = split(/\t/);
	}
	else {
	    gen_head();
	    if($inignore) {
	    }
	    elsif($incenter > 0) {
		$incenter--;
		if($incenter == 0) {
		    print OUT "</CENTER>\n";
		}
	    }
	    elsif($indisp eq "L") {
		printf OUT ("%s<br/>\n", $_);
	    }
	    else {
		printf OUT ("%s\n", $_);
	    }
	}
    }
    close(IN);
    set_font("R", $lastfont);
    print OUT <<EOF;
</body>
</html>
EOF
    close(OUT);

my $x = <<EOF;
.TS
center allbox;
c s s
l r r
l n n.
Early R90/6 (to 9/75)
Part	diameter	area
caliper	38mm	1134.0 mm\[u00B2]
master	14mm	153.9 mm\[u00B2]
ratio		7.37
.TE
EOF

sub gen_head () {
my $ltitle, $lsubtitle;
    $ltitle = $main::title || $main::trfile;
    $lsubtitle = $main::subtitle || "";
    if(!$head) {
	print OUT <<EOF;
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
<meta name="viewport" content="width=device-width, initial-scale=1">
<title>$ltitle</title>
</head>
<body>
<style type="text/css">
.courier {
    font-family: Courier, monospaced;
}
.helvetica {
    font-family: Arial, Helvetica, Verdana, sans-serif;
}
</style>
<center>
<h4>$ltitle</h4>
<h5>$lsubtitle</h5>
</center>
EOF
	$head = 1;
    }
}

my $isbold, $isital, $ishelv, $isrom, $iscour;
my $wasbold, $wasital, $washelv, $wasrom, $wascour;
sub set_font ($$) {
my($currfont, $lastfont)=@_;
my $realfont;
    $wasrom = $isrom;
    $washelv = $ishelv;
    $wascour = $iscour;
    $wasbold = $isbold;
    $wasital = $isital;
    if($currfont eq "H") {
	$isrom  = 0;
	$ishelv = 1;
	$iscour = 0;
	$isbold = 0;
	$isital = 0;
    }
    elsif($currfont eq "R") {
	$isrom  = 1;
	$ishelv = 0;
	$iscour = 0;
	$isbold = 0;
	$isital = 0;
    }
    elsif($currfont eq "CW") {
	$isrom  = 0;
	$ishelv = 0;
	$iscour = 1;
	$isbold = 0;
	$isital = 0;
    }
    elsif($currfont eq "HB") {
	$isrom  = 0;
	$ishelv = 1;
	$iscour = 0;
	$isbold = 1;
	$isital = 0;
    }
    elsif($currfont eq "RB") {
	$isrom  = 1;
	$ishelv = 0;
	$iscour = 0;
	$isbold = 1;
	$isital = 0;
    }
    elsif($currfont eq "HI") {
	$isrom  = 0;
	$ishelv = 1;
	$iscour = 0;
	$isital = 1;
	$isbold = 0;
    }
    elsif($currfont eq "RI") {
	$isrom  = 1;
	$ishelv = 0;
	$iscour = 0;
	$isital = 1;
	$isbold = 0;
    }
    elsif($currfont eq "I") {
	$isrom  = 1;
	$ishelv = 0;
	$iscour = 0;
	$isital = 1;
	$isbold = 0;
    }
    elsif($currfont eq "B") {
	$isrom  = 1;
	$ishelv = 0;
	$iscour = 0;
	$isital = 0;
	$isbold = 1;
    }
    $realfont = "";
    if($isrom ) { $realfont .= "R"; }
    if($ishelv) { $realfont .= "H"; } elsif($washelv) { $realfont .= "h"; }
    if($iscour) { $realfont .= "C"; } elsif($wascour) { $realfont .= "c"; }
    if($isital) { $realfont .= "I"; } elsif($wasital) { $realfont .= "i"; }
    if($isbold) { $realfont .= "B"; } elsif($wasbold) { $realfont .= "b"; }
    $main::debug{'f'} && printf("font set to from %s to %s, realfont = %s\n", $lastfont, $currfont, $realfont);
    if($realfont =~ /R/) {
	# print OUT "<I>";
    }
    elsif($realfont =~ /r/) {
	# print OUT "</I>";
    }
    if($realfont =~ /c/) {
	print OUT "</span>";
    }
    if($realfont =~ /h/) {
	print OUT "</span>";
    }
    if($realfont =~ /b/) {
	print OUT "</B>";
    }
    if($realfont =~ /i/) {
	print OUT "</I>";
    }
    if($realfont =~ /H/) {
	print OUT "<span class=\"helvetica\">";
    }
    if($realfont =~ /C/) {
	print OUT "<span class=\"courier\">";
    }
    if($realfont =~ /B/) {
	print OUT "<B>";
    }
    if($realfont =~ /I/) {
	print OUT "<I>";
    }
}

sub usage () {
    printf("Usage: tr2html.pl [-dtable] [-dfont] [-dheader] filename.nr\n");
    exit(-1);
}
